
// all_of, any_of, none_of, copy, copy_if, count_if
// source: cppreference.com

#include <vector>
#include <numeric>
#include <ranges>
#include <algorithm>
#include <iterator>             // ostream_iterator
#include <iostream>
#include <functional>

int main()
{
    std::vector<int> v(10, 2);
    std::partial_sum(v.cbegin(), v.cend(), v.begin());

    std::cout << "Among the numbers: ";
    std::ranges::copy(v, std::ostream_iterator<int>(std::cout, " "));
    std::cout << '\n';

    if (std::ranges::all_of(v, [](int i){ return i % 2 == 0; })) {
        std::cout << "All numbers are even\n";
    }
    if (std::ranges::none_of(v, [](int i){ return i % 2 == 1; })) {
        std::cout << "None of them are odd\n";
    }

    struct DivisibleBy {
        const int d;
        DivisibleBy(int n) : d(n) {}
        bool operator()(int n) const { return n % d == 0; }
    };

    if (std::ranges::any_of(v, DivisibleBy(7))) {
        std::cout << "At least one number is divisible by 7\n";
    }
    std::cout << "numbers divisible by 5: ";
    std::ranges::copy_if(v, std::ostream_iterator<int>(std::cout, " "),
                         DivisibleBy(5));
    std::cout << "\t(count = " <<
                 std::ranges::count_if(v, DivisibleBy(5));
    std::cout << ")\n\n";
}

