
// example of array creation+initialization function templete

#include <array>
#include <iostream>

// function: fixed number of int elements
std::array<int,3> getval(int a, int b, int c) {
    return {a,b,c};     // type deduced from return type
}

// template: fixed number of elements of the same type
template<typename T>
std::array<T,3> getval0(T a, T b, T c) {
    return {a,b,c};     // type deduced
}

// variadic template: any number of elements
template<typename T, typename ... Types>
std::array<T,1+sizeof...(Types)> getval1(T a, Types ... params) {
    return {a,params...};
}

// variadic template: C++17, CTAD="Class Template Argument Deduction" (no need for <>)
template<typename ... Types>
auto getval2(const Types & ... params) {
    return std::array{params...};       // deduce type and size of array
}

// variadic template: C++20, auto
auto getval3(const auto & ... params) { // pass by reference, read-only
    return std::array{params...};       // deduce type and size, copy
}

auto getval3a(auto ... params) {        // pass by value (const possible)
    return std::array{params...};       // deduce type and size, copy
}

auto getval3b(auto && ... params) {     // universal/forwarding reference
    return std::array{params...};       // deduce type and size, copy/move ??
}

// variadic template: C++20, auto, more efficient (no copy)
auto getval4(auto && ... params) {
    return std::array{std::forward<decltype(params)>(params)...};
}

//////////////////////////////////////////////////////////////////

void print(const auto &a) {
    for(const auto &x : a) {
        std::cout << x << " ";
    }
}

int main() {
    auto a{getval4(1,2,3)};
    print(a);
    std::cout << "\n";
    return a[1];
}
