
// C++ namespace example
// GOAL: Explain how ADL and "using directive" works
// Try to uncomment lines marked #1, #2, #3
// N4944: 6.5.3 Unqualified name lookup [basic.lookup.unqual]
//        9.8.4 Using namespace directive [namespace.udir]
//////////////////////////////////////////////////////////////////

/* Hierarchy of namespaces:

   :: - X - Y - Z
         \
          A
*/

#include <iostream>

void f(int x)  { std::cout << "::f(int)\n"; }
void f()       { std::cout << "::f()\n"; }

namespace E {
    class C {};
    void f(C) { std::cout << "E::f(E::C)\n"; }
}

namespace X {
//    void f(int x) { std::cout << "X::f(int)\n"; } //#1 ambiguity
    namespace Y {
        void f(int x) { std::cout << "X::Y::f(int)\n"; }
        namespace Z {
            void f(int x) { std::cout << "X::Y::Z::f(int)\n"; }
        }
    } // namespace X::Y

    namespace A {
        /// uncommenting next 2 lines behave differently
//        void f(int x) { std::cout << "X::A::f(int)\n"; } //#2
        void test1() {
            using namespace Y::Z; /// relative path to X::Y::Z
            /// X::Y::Z::f(int) overloads like to be declared in X
            /// = first common namespace for X::Y::Z and X::A::g
//            using namespace Y;          //#3 add X::Y::f(int)

            f(1); // unqualified name used, int argument  ==> normal search
//            f();      //#4
            ::f();      // fully qualified = no problem

            using namespace Y;          // add X::Y::f(int)
            using namespace Y;          // no problem
            /// no f() call here ==> no ambiguity...
//            f(1);     //#5 ambiguity Y Z
//            extern void f(int x);  f(1);  //#9 no X::A::f(int) defined
        }
        void test2(E::C c) {
            f(c);       // unqualified name + argument ==> ADL search
//            (f)(c);   //#6 parentheses ==> no ADL, not found
        }
    } // namespace X::A
}

int main() {
    X::A::test1(); // fully qualified names
    E::C c;
    X::A::test2(c);
}

