
// namespaces and lookup
/*
// Q: where using directive "places" symbols?
// A: C++ standard 7.3.4
A using-directive specifies that the names in the nominated namespace can be
used in the scope in which the using-directive appears after the
using-directive. During unqualified name lookup (3.4.1), the names appear as
if they were declared in the nearest enclosing namespace which contains both
the using-directive and the nominated namespace. [ Note: In this context,
"contains" means "contains directly or indirectly". -- end note ]
*/

#include <iostream>
#include <typeinfo>
using namespace std;

// will not be used
void f(long i) {
    cout << "::f("<< i << ")\t T=" << typeid(i).name() << endl;
}

namespace X {
    namespace A {
        void f(int i) {
            cout << "A::f("<< i << ")\t T=" << typeid(i).name() << endl;
        }
    }

    namespace B {
        void f(double d) {
            cout << "A::f("<< d << ")\t T=" << typeid(d).name() << endl;
        }
    }

    void f() {
        using namespace A;
        using namespace B;
        // info: ADL not possible, using unqualified lookup
        f(1);    // f(int) found as if declared in namespace X
        f(3.14); // f(double) found in namespace X
#if defined(ERR)
        f(1L);   // no f(long) in namespace X, search does not continue
        /*
            In function ‘void X::f()’:
            error: call of overloaded ‘f(long int)’ is ambiguous
            note: candidates are: void X::B::f(double)
            note:                 void X::A::f(int)
        */
#endif
    }

} // namespace X

int main() {
    X::f();
}

