
// http://en.cppreference.com/w/cpp/thread/thread/join
// needs linking with -lpthread
// Warning: cout in threads can mangle output

#include <iostream>
#include <thread>
#include <chrono>

using std::cout;
using std::endl;

void foo() {
    cout << "starting hepler thread...\n";
    // simulate expensive operation
    std::this_thread::sleep_for(std::chrono::seconds(2));
    cout << "ending hepler thread...\n";
}

void bar() {
    cout << "starting another hepler thread...\n";
    // simulate expensive operation
    std::this_thread::sleep_for(std::chrono::seconds(3));
    cout << "ending another hepler thread...\n";
}

int main() {
    unsigned int n = std::thread::hardware_concurrency();
    cout << n << " concurrent threads are supported.\n";
    cout << "main thread id="<< std::this_thread::get_id() << "\n";
    cout << endl;

    std::thread helper1(foo);
    std::thread helper2(bar);
    std::thread helper3(foo);

    // main thread work
    std::this_thread::sleep_for(std::chrono::seconds(1));

    cout << "waiting for helper1 thread to finish..." << endl;
    helper1.join();
    cout << "waiting for helper2 thread to finish..." << endl;
    helper2.join();
#if 1
    cout << "waiting for helper3 thread to finish..." << endl;
    helper3.join();   // exception if missing
#endif

    cout << endl;
    cout << "done!\n";
}

